from abc import ABC, abstractmethod
from typing import Dict, Union, List, Optional
from pathlib import Path
from pyroaring import BitMap

class BaseIndex(ABC):
    """Base class for all indexes"""
    
    @abstractmethod
    def __init__(self, 
                 index_dir: Optional[Path] = None,
                 max_chinese_length: int = 4,
                 min_term_length: int = 2,
                 buffer_size: int = 100000,
                 shard_bits: int = 8,
                 cache_size: int = 1000):
        """
        Initialize the index
        
        Args:
            index_dir: The directory to store the index files
            max_chinese_length: The maximum length of Chinese substrings
            min_term_length: The minimum length of terms
            buffer_size: The size of the memory buffer
            shard_bits: The number of bits for the shard
            cache_size: The size of the cache
        """
        pass

    @abstractmethod
    def add_terms(self, doc_id: int, terms: Dict[str, Union[str, int, float]]) -> None:
        """Add terms to the index"""
        pass

    @abstractmethod
    def search(self, query: str, enable_fuzzy: bool = False, min_results: int = 5, score_threshold: Optional[float] = None) -> Union[BitMap, List[tuple[int, float]]]:
        """Search for a query"""
        pass

    @abstractmethod
    def remove_document(self, doc_id: int) -> None:
        """Remove a document from the index"""
        pass

    @abstractmethod
    def merge_buffer(self) -> None:
        """Merge the buffer"""
        pass

    @abstractmethod
    def save(self, incremental: bool = True) -> None:
        """Save the index"""
        pass

    @abstractmethod
    def load(self) -> bool:
        """Load the index"""
        pass

    @abstractmethod
    def build_word_index(self) -> None:
        """Build the word index"""
        pass 


from enum import Enum
from pathlib import Path
from typing import Optional

from .inverted import InvertedIndex


class IndexType(Enum):
    """Index type enum"""
    INVERTED = "inverted"  # Inverted index


class IndexFactory:
    """Index factory class"""
    
    @staticmethod
    def create_index(index_type: IndexType,
                    index_dir: Optional[str] = None,
                    max_chinese_length: int = 4,
                    min_term_length: int = 2,
                    buffer_size: int = 100000,
                    shard_bits: int = 8,
                    cache_size: int = 1000,
                    fuzzy_threshold: float = 0.7,
                    fuzzy_max_distance: int = 2,
                    **kwargs) -> InvertedIndex:
        """
        Create an index instance
        
        Args:
            index_type: The index type
            index_dir: The directory to store the index files
            max_chinese_length: The maximum length of Chinese substrings
            min_term_length: The minimum length of terms
            buffer_size: The size of the memory buffer
            shard_bits: The number of bits for the shard
            cache_size: The size of the cache
            fuzzy_threshold: The similarity threshold for fuzzy search (0.0-1.0)
            fuzzy_max_distance: The maximum edit distance for fuzzy search
            **kwargs: Other parameters
        """
        index_dir_path = Path(index_dir) if index_dir else None
        
        if index_type == IndexType.INVERTED:
            return InvertedIndex(
                index_dir=index_dir_path,
                max_chinese_length=max_chinese_length,
                min_term_length=min_term_length,
                buffer_size=buffer_size,
                shard_bits=shard_bits,
                cache_size=cache_size,
                fuzzy_threshold=fuzzy_threshold,
                fuzzy_max_distance=fuzzy_max_distance
            )
        else:
            raise ValueError(f"Unsupported index type: {index_type}") 


from concurrent.futures import ThreadPoolExecutor
from typing import List, Dict, Union
import numpy as np

from .base import BaseIndex

class DocumentInserter:
    """Document inserter, responsible for handling bulk document insertion"""
    
    def __init__(self, 
                 index: BaseIndex,
                 num_workers: int = 8,
                 batch_size: int = 10000,
                 shard_size: int = 500_000):
        """
        Initialize the document inserter
        
        Args:
            index (BaseIndex): The index instance
            num_workers (int): The number of worker threads for parallel processing
            batch_size (int): The batch size
            shard_size (int): The shard size
        """
        self.index = index
        self.num_workers = num_workers
        self.batch_size = batch_size
        self.shard_size = shard_size
        self._batch_count = 0

    def add_documents(self, 
                     doc_ids: Union[int, List[int]], 
                     docs: Union[Dict[str, Union[str, int, float]], 
                               List[Dict[str, Union[str, int, float]]]]) -> None:
        """
        Bulk add documents
        
        Args:
            doc_ids (int | list): The document ID or ID list
            docs (dict | list): The document content or document list
        """
        # Standardize the input
        if isinstance(doc_ids, int):
            doc_ids = [doc_ids]
            docs = [docs] if isinstance(docs, dict) else docs
        else:
            docs = docs if isinstance(docs, list) else [docs]
        
        if len(doc_ids) != len(docs):
            raise ValueError("The length of the document ID list and the document list must be the same")
        
        # Process large batches of documents in parallel
        total_docs = len(docs)
        chunk_size = max(10000, total_docs // (self.num_workers * 2))
        
        def process_chunk(start_idx: int, chunk_docs: List[dict]) -> None:
            for i, doc in enumerate(chunk_docs):
                self.index.add_terms(doc_ids[start_idx + i], doc)
        
        if total_docs < chunk_size:
            process_chunk(0, docs)
        else:
            chunks = []
            for i in range(0, total_docs, chunk_size):
                chunks.append((i, docs[i:i + chunk_size]))
            
            with ThreadPoolExecutor(max_workers=self.num_workers) as executor:
                executor.map(lambda x: process_chunk(*x), chunks)
        
        self._batch_count += len(docs)
        
        # Check if merging and saving is needed
        if len(self.index.index_buffer) >= self.index.buffer_size * 4:
            self.index.merge_buffer()
        
        if self._batch_count >= self.batch_size * 4:
            self.flush()

    def update_document(self, 
                       doc_id: int, 
                       fields: Union[Dict[str, Union[str, int, float]], 
                       List[Dict[str, Union[str, int, float]]]]):
        """Update a document in the index
        
        Args:
            doc_id: The document ID to update
            fields: The fields to update
        """
        self.index.update_terms(doc_id, fields)

    def batch_update_document(self, 
                             doc_ids: List[int], 
                             fields: List[Dict[str, Union[str, int, float]]]):
        """批量更新多个文档
        
        Args:
            doc_ids: 文档ID列表
            fields: 文档字段列表，与doc_ids一一对应
        """
        if len(doc_ids) != len(fields):
            raise ValueError("文档ID列表和文档字段列表长度不匹配")
            
        self.index.batch_update_terms(doc_ids, fields)

    def batch_remove_document(self, doc_ids: List[int]):
        """批量删除多个文档
        
        Args:
            doc_ids: 要删除的文档ID列表
        """
        self.index.batch_remove_document(doc_ids)
        if self.index.index_dir:
            self.index.save(incremental=True)

    def flush(self) -> None:
        """Flush the buffer and save"""
        self.index.merge_buffer()
        if self._batch_count > 0:
            self.index.build_word_index()
            if self.index.index_dir:
                shard_id = self._batch_count // self.shard_size
                self.index.save(incremental=True)
            self._batch_count = 0

    def from_pandas(self, df, id_column=None, text_columns=None):
        """Import data from a pandas DataFrame
        
        Args:
            df (pandas.DataFrame): The pandas DataFrame to import
            id_column (str): The name of the document ID column, if None then use the row index
            text_columns (list): The list of text columns to index, if None then use all string columns
        """
        try:
            import pandas as pd
        except ImportError:
            raise ImportError("Use from_pandas requires installing pandas, please execute: pip install nanofts[pandas]")

        if not isinstance(df, pd.DataFrame):
            raise TypeError("df must be a pandas DataFrame object")

        # Get the document ID
        if id_column is None:
            doc_ids = df.index.tolist()
        else:
            doc_ids = df[id_column].tolist()

        # Get the columns to index
        if text_columns is None:
            text_columns = df.select_dtypes(include=['object', 'string']).columns.tolist()
        
        # Build the document list
        docs = []
        for _, row in df[text_columns].iterrows():
            doc = {col: str(val) for col, val in row.items() if pd.notna(val)}
            docs.append(doc)

        # Add documents and flush
        self.add_documents(doc_ids, docs)
        self.flush()

    def from_polars(self, df, id_column=None, text_columns=None):
        """Import data from a polars DataFrame
        
        Args:
            df (polars.DataFrame): The polars DataFrame to import
            id_column (str): The name of the document ID column, if None then use the row index
            text_columns (list): The list of text columns to index, if None then use all string columns
        """
        try:
            import polars as pl
        except ImportError:
            raise ImportError("Use from_polars requires installing polars, please execute: pip install nanofts[polars]")

        if not isinstance(df, pl.DataFrame):
            raise TypeError("df must be a polars DataFrame object")

        # Get the document ID
        if id_column is None:
            doc_ids = list(range(len(df)))
        else:
            doc_ids = df[id_column].to_list()

        # Get the columns to index
        if text_columns is None:
            text_columns = [col for col in df.columns if df[col].dtype == pl.Utf8]

        # Build the document list
        docs = []
        for row in df.select(text_columns).iter_rows():
            doc = {col: str(val) for col, val in zip(text_columns, row) if val is not None}
            docs.append(doc)

        # Add documents and flush
        self.add_documents(doc_ids, docs)
        self.flush() 
    
    def from_arrow(self, table, id_column=None, text_columns=None):
        """Import data from an arrow table
        
        Args:
            table (pyarrow.Table): The pyarrow Table object
            id_column (str): The name of the document ID column, if None then use the row index
            text_columns (list): The list of text columns to index, if None then use all string columns
        """
        try:
            import pyarrow as pa
        except ImportError:
            raise ImportError("Use from_arrow requires installing pyarrow, please execute: pip install nanofts[pyarrow]")

        if not isinstance(table, pa.Table):
            raise TypeError("table must be a pyarrow Table object")

        # Get the columns to index
        if text_columns is None:
            text_columns = [field.name for field in table.schema 
                           if pa.types.is_string(field.type)]
            if id_column and id_column in text_columns:
                text_columns.remove(id_column)

        # Get the document ID
        if id_column is None:
            doc_ids = range(table.num_rows)
        else:
            # Fix: correctly handle the id column
            id_array = table.column(id_column)
            if pa.types.is_integer(id_array.type):
                doc_ids = id_array.to_numpy()
            else:
                # If not an integer type, try to convert
                doc_ids = [int(val.as_py()) for val in id_array]

        # Optimize: use batch processing to handle large tables
        batch_size = 50000
        num_batches = (table.num_rows + batch_size - 1) // batch_size

        for batch_idx in range(num_batches):
            start_idx = batch_idx * batch_size
            end_idx = min(start_idx + batch_size, table.num_rows)
            
            # Use slice to avoid full copy
            batch = table.slice(start_idx, end_idx - start_idx)
            
            # Use Arrow's column access directly
            docs = []
            for i in range(batch.num_rows):
                doc = {}
                for col in text_columns:
                    val = batch.column(col)[i].as_py()
                    if val is not None:
                        doc[col] = str(val)
                docs.append(doc)
            
            # Add this batch of documents
            batch_doc_ids = (doc_ids[start_idx:end_idx] if isinstance(doc_ids, (list, np.ndarray)) 
                            else range(start_idx, end_idx))
            self.add_documents(list(batch_doc_ids), docs)  # Ensure converted to list

        # Flush at the end
        self.flush()

    def from_parquet(self, path, id_column=None, text_columns=None):
        """Import data from a parquet file
        
        Args:
            path (str): The path to the parquet file
            id_column (str): The name of the document ID column, if None then use the row index
            text_columns (list): The list of text columns to index, if None then use all string columns
        """
        try:
            import pyarrow.parquet as pq
        except ImportError:
            raise ImportError("Use from_parquet requires installing pyarrow, please execute: pip install nanofts[pyarrow]")

        # Optimize: use memory mapping to read large files
        table = pq.read_table(path, memory_map=True)
        self.from_arrow(table, id_column, text_columns)

    def from_csv(self, path, id_column=None, text_columns=None):
        """Import data from a csv file
        
        Args:
            path (str): The path to the csv file
            id_column (str): The name of the document ID column, if None then use the row index
            text_columns (list): The list of text columns to index, if None then use all string columns
        """
        try:
            import polars as pl
        except ImportError:
            raise ImportError("Use from_csv requires installing polars, please execute: pip install nanofts[polars]")

        df = pl.read_csv(path)
        self.from_polars(df, id_column, text_columns)  
        


import pytest

from nanofts import FullTextSearch

@pytest.fixture
def test_data():
    """Test data fixture"""
    return [
        {"title": "Hello World", "content": "Python 全文搜索器"},  # id: 0
        {"title": "GitHub Copilot", "content": "代码自动生成"},   # id: 1
        {"title": "全文搜索", "content": "支持多语言", "tags": "测试数据"},  # id: 2
        {"title": "hello", "content": "WORLD", "number": 123},  # id: 3
        {"title": "数据处理", "content": "搜索引擎"},  # id: 4
        {"title": "hello world", "content": "示例文本"},  # id: 5
        {"title": "混合文本", "content": "Mixed 全文内容测试"},  # id: 6
        {"title": "hello world 你好", "content": "示例文本"},  # id: 7
    ]

@pytest.fixture
def fts(tmp_path):
    """Create a FTS instance fixture"""
    index_dir = tmp_path / "fts_index"
    return FullTextSearch(
        index_dir=str(index_dir),
        batch_size=1000,
        buffer_size=5000,
        drop_if_exists=True
    )

def test_basic_search(fts, test_data):
    """Test the basic search functionality"""
    # Add test data
    doc_ids = list(range(len(test_data)))
    fts.add_document(doc_ids, test_data)
    fts.flush()

    # Test exact matching
    test_cases = [
        ("Hello World", [0, 5]),  # Only match the complete "Hello World"
        ("hello world", [0, 5]),  # Case insensitive
        ("全文搜索", [0, 2]),  # Modified: Include all documents containing these words
        ("mixed", [6])
    ]

    for query, expected in test_cases:
        assert sorted(fts.search(query)) == expected

def test_chinese_search(fts, test_data):
    """Test the Chinese search functionality"""
    # Add test data
    doc_ids = list(range(len(test_data)))
    fts.add_document(doc_ids, test_data)
    fts.flush()

    test_cases = [
        ("全文", [0, 2, 6]),
        ("搜索", [0, 2, 4]),
        ("测试", [2, 6])
    ]

    for query, expected in test_cases:
        assert sorted(fts.search(query)) == expected

def test_phrase_search(fts, test_data):
    """Test the phrase search functionality"""
    # Add test data
    doc_ids = list(range(len(test_data)))
    fts.add_document(doc_ids, test_data)
    fts.flush()

    test_cases = [
        ("hello world", [0, 5]),  # Only match the complete phrase
        ("全文 搜索", [0, 2]),
        ("python 搜索", [0])
    ]

    for query, expected in test_cases:
        assert sorted(fts.search(query)) == expected

def test_incremental_update(fts, test_data):
    """Test the incremental update functionality"""
    # Add initial test data
    doc_ids = list(range(len(test_data)))
    fts.add_document(doc_ids, test_data)
    fts.flush()

    # Add new document
    new_doc = {"title": "新增文档", "content": "测试全文搜索", "tags": "hello world test"}
    new_doc_id = len(test_data)
    fts.add_document(new_doc_id, new_doc)
    fts.flush()

    # Test if the new document is searchable
    test_cases = [
        ("新增", [new_doc_id]),
        ("测试", [2, 6, new_doc_id]),
        ("hello world", [0, 5])  # phrase search doesn't match "hello world test"
    ]

    for query, expected in test_cases:
        assert sorted(fts.search(query)) == expected

    # Test deleting documents
    fts.remove_document(new_doc_id)
    fts.flush()  # Ensure deletion is persisted, especially on Windows
    
    for query, expected in test_cases:
        result = sorted(fts.search(query))
        # Remove the deleted document ID from the expected result
        expected = [id for id in expected if id != new_doc_id]
        assert result == expected

def test_index_persistence(fts, test_data, tmp_path):
    """Test the index persistence functionality"""
    # Add test data
    doc_ids = list(range(len(test_data)))
    fts.add_document(doc_ids, test_data)
    fts.flush()

    # Create a new FTS instance to load the existing index
    index_dir = tmp_path / "fts_index"
    fts_reload = FullTextSearch(index_dir=str(index_dir))

    test_cases = [
        ("hello world", [0, 5]),  # Only match the complete phrase
        ("全文", [0, 2, 6]),
        ("搜索", [0, 2, 4])
    ]

    for query, expected in test_cases:
        assert sorted(fts_reload.search(query)) == expected

def test_empty_search(fts):
    """Test the empty index search"""
    assert len(fts.search("任意查询")) == 0

def test_invalid_document_input(fts):
    """Test the invalid document input"""
    with pytest.raises(ValueError):
        fts.add_document([1, 2], [{"title": "doc1"}])  # The document ID and document number do not match

def test_case_insensitive_search(fts, test_data):
    """Test the case insensitive search"""
    doc_ids = list(range(len(test_data)))
    fts.add_document(doc_ids, test_data)
    fts.flush()

    # Test that queries with different cases return the same results
    lower_case = sorted(fts.search("hello world"))
    upper_case = sorted(fts.search("HELLO WORLD"))
    mixed_case = sorted(fts.search("Hello World"))

    assert lower_case == upper_case == mixed_case 

def test_from_pandas(tmp_path):
    """Test importing data from a pandas DataFrame"""
    pd = pytest.importorskip("pandas")
    
    # Create test data
    df = pd.DataFrame({
        'id': [1, 2, 3],
        'title': ['Hello World', '全文搜索', 'Test Document'],
        'content': ['This is a test', '支持多语言', 'Another test'],
        'tags': ['test, hello', '搜索, 测试', 'doc, test']
    })
    
    # Create an instance and import data
    fts = FullTextSearch(index_dir=str(tmp_path / "fts_index"))
    fts.from_pandas(df, id_column='id')
    
    # Test the search results
    assert sorted(fts.search("test")) == [1, 3]
    assert sorted(fts.search("搜索")) == [2]
    assert sorted(fts.search("hello")) == [1]

def test_from_polars(tmp_path):
    """Test importing data from a polars DataFrame"""
    pl = pytest.importorskip("polars")
    
    # Create test data
    df = pl.DataFrame({
        'id': [1, 2, 3],
        'title': ['Hello World', '全文搜索', 'Test Document'],
        'content': ['This is a test', '支持多语言', 'Another test'],
        'tags': ['test, hello', '搜索, 测试', 'doc, test']
    })
    
    # Create an instance and import data
    fts = FullTextSearch(index_dir=str(tmp_path / "fts_index"))
    fts.from_polars(df, id_column='id')
    
    # Test the search results
    assert sorted(fts.search("test")) == [1, 3]
    assert sorted(fts.search("搜索")) == [2]
    assert sorted(fts.search("hello")) == [1]

def test_from_arrow(tmp_path):
    """Test importing data from a pyarrow Table"""
    pa = pytest.importorskip("pyarrow")
    
    # Create test data
    data = {
        'id': [1, 2, 3],
        'title': ['Hello World', '全文搜索', 'Test Document'],
        'content': ['This is a test', '支持多语言', 'Another test'],
        'tags': ['test, hello', '搜索, 测试', 'doc, test']
    }
    table = pa.Table.from_pydict(data)
    
    # Create an instance and import data
    fts = FullTextSearch(index_dir=str(tmp_path / "fts_index"))
    fts.from_arrow(table, id_column='id')
    
    # Test the search results
    assert sorted(fts.search("test")) == [1, 3]
    assert sorted(fts.search("搜索")) == [2]
    assert sorted(fts.search("hello")) == [1]

def test_from_parquet(tmp_path):
    """Test importing data from a parquet file"""
    pa = pytest.importorskip("pyarrow")
    pq = pytest.importorskip("pyarrow.parquet")
    
    # Create test data
    data = {
        'id': [1, 2, 3],
        'title': ['Hello World', '全文搜索', 'Test Document'],
        'content': ['This is a test', '支持多语言', 'Another test'],
        'tags': ['test, hello', '搜索, 测试', 'doc, test']
    }
    table = pa.Table.from_pydict(data)
    
    # Save as a parquet file
    parquet_path = tmp_path / "test.parquet"
    pq.write_table(table, parquet_path)
    
    # Create an instance and import data
    fts = FullTextSearch(index_dir=str(tmp_path / "fts_index"))
    fts.from_parquet(parquet_path, id_column='id')
    
    # Test the search results
    assert sorted(fts.search("test")) == [1, 3]
    assert sorted(fts.search("搜索")) == [2]
    assert sorted(fts.search("hello")) == [1] 

def test_from_csv(tmp_path):
    """Test importing data from a csv file"""
    pd = pytest.importorskip("pandas")
    
    # Create test data
    df = pd.DataFrame({
        'id': [1, 2, 3],
        'title': ['Hello World', '全文搜索', 'Test Document'],
        'content': ['This is a test', '支持多语言', 'Another test'],
        'tags': ['test, hello', '搜索, 测试', 'doc, test']
    })
    
    # Save as a csv file
    csv_path = tmp_path / "test.csv"
    df.to_csv(csv_path, index=False)
    
    # Create an instance and import data
    fts = FullTextSearch(index_dir=str(tmp_path / "fts_index"))
    fts.from_csv(csv_path, id_column='id')

    # Test the search results
    assert sorted(fts.search("test")) == [1, 3]
    assert sorted(fts.search("搜索")) == [2]
    assert sorted(fts.search("hello")) == [1]   

def test_update_document(fts, test_data):
    """Test updating document terms"""
    # Add initial test data
    doc_ids = list(range(len(test_data)))
    fts.add_document(doc_ids, test_data)
    fts.flush()

    # Initial search to verify setup
    assert sorted(fts.search("hello world")) == [0, 5]
    assert sorted(fts.search("全文搜索")) == [0, 2]

    # Update document 0
    updated_doc = {
        "title": "Updated Title",
        "content": "新的内容 New Content",
        "tags": "updated"
    }
    fts.update_document(0, updated_doc)
    fts.flush()

    # Verify old terms are removed
    assert sorted(fts.search("hello world")) == [5]  # Document 0 no longer matches
    assert sorted(fts.search("全文搜索")) == [2]  # Document 0 no longer matches

    # Verify new terms are searchable
    assert sorted(fts.search("updated")) == [0]
    assert sorted(fts.search("新的内容")) == [0]

    # Update document with mixed content
    mixed_doc = {
        "title": "Mixed 混合",
        "content": "Content 内容",
        "tags": "test"
    }
    fts.update_document(5, mixed_doc)
    fts.flush()

    # Verify the updates
    assert set(fts.search("mixed")) == {5, 6}  # Both doc 5 and 6 contain "mixed"
    assert set(fts.search("混合")) == {5, 6}  # Both doc 5 and 6 contain "混合"
    assert set(fts.search("hello world")) == {7}  # Only document 7 still contains "hello world" phrase

def test_update_nonexistent_document(fts):
    """Test updating a document that doesn't exist"""
    doc = {
        "title": "New Doc",
        "content": "Some content"
    }
    # Should not raise an error
    fts.update_document(999, doc)
    fts.flush()
    
    # Verify the document is added
    assert sorted(fts.search("new doc")) == [999]

def test_batch_update_document(fts, test_data):
    """测试批量更新文档功能"""
    # 添加初始测试数据
    doc_ids = list(range(len(test_data)))
    fts.add_document(doc_ids, test_data)
    fts.flush()
    
    # 初始搜索验证
    assert sorted(fts.search("hello world")) == [0, 5]
    assert sorted(fts.search("全文搜索")) == [0, 2]
    assert sorted(fts.search("github")) == [1]
    
    # 批量更新文档
    updated_docs = [
        {
            "title": "Updated Title 1",
            "content": "新的内容 New Content 1",
            "tags": "tag1"
        },
        {
            "title": "Updated GitHub",
            "content": "更新的代码生成器",
            "tags": "ai"
        },
        {
            "title": "Mixed 混合",
            "content": "Content 内容",
            "tags": "test"
        }
    ]
    
    # 更新文档 0, 1, 5
    fts.update_document([0, 1, 5], updated_docs)
    fts.flush()

    # 验证旧词条已被删除
    assert sorted(fts.search("hello world")) == [7]  # 文档 0 和 5 已更新，只有文档 7 仍然包含短语 "hello world"
    assert sorted(fts.search("全文搜索")) == [2]  # 文档 0 不再匹配
    assert sorted(fts.search("github copilot")) == []  # 文档 1 不再匹配
    
    # 验证新词条可搜索
    assert sorted(fts.search("tag1")) == [0]  # 只有文档 0 包含 tag1
    assert sorted(fts.search("新的内容")) == [0]
    assert sorted(fts.search("更新的代码")) == [1]
    assert sorted(fts.search("混合")) == [5, 6]  # 文档 5 和 6 都包含"混合"
    
    # 验证未更新的文档仍然可搜索
    assert sorted(fts.search("测试数据")) == [2]
    assert sorted(fts.search("world")) == [7]  # 只有文档 7 仍然包含 "world"
    
    # 测试批量更新的性能优势 - 同时更新多个文档
    large_batch = []
    large_ids = list(range(100, 150))
    
    for i in range(50):
        large_batch.append({
            "title": f"Batch Doc {i}",
            "content": f"批量文档内容 {i}",
            "tags": f"tag{i}"
        })
    
    # 批量添加
    fts.add_document(large_ids, large_batch)
    fts.flush()
    
    # 验证添加成功
    assert len(fts.search("批量文档内容")) == 50
    
    # 批量更新
    updated_batch = []
    for i in range(50):
        updated_batch.append({
            "title": f"Updated Batch {i}",
            "content": f"更新的批量内容 {i}",
            "tags": f"updated{i}"
        })
    
    fts.update_document(large_ids, updated_batch)
    fts.flush()

    # 验证更新成功 - 使用更具体的搜索词
    # 验证旧内容已被删除 - 搜索原来的tag
    assert len(fts.search("tag0")) == 0  # 特定文档tag应该被删除
    assert len(fts.search("tag10")) == 0  # 特定文档tag应该被删除  
    assert len(fts.search("tag49")) == 0  # 特定文档tag应该被删除
    
    # 验证新内容已添加
    assert len(fts.search("更新的批量内容")) == 50  # 新内容已添加

def test_batch_remove_document(fts, test_data):
    """测试批量删除文档功能"""
    # 添加初始测试数据
    doc_ids = list(range(len(test_data)))
    fts.add_document(doc_ids, test_data)
    fts.flush()
    
    # 初始搜索验证
    assert sorted(fts.search("hello world")) == [0, 5]
    assert sorted(fts.search("全文搜索")) == [0, 2]
    assert sorted(fts.search("github")) == [1]
    
    # 批量删除文档 0, 1, 5
    fts.remove_document([0, 1, 5])
    fts.flush()

    # 验证文档已被删除
    assert sorted(fts.search("hello world")) == [7]  # 文档 0 和 5 已删除，只有文档 7 仍然包含短语 "hello world"
    assert sorted(fts.search("全文搜索")) == [2]  # 文档 0 已删除
    assert sorted(fts.search("github")) == []  # 文档 1 已删除
    
    # 验证未删除的文档仍然可搜索
    assert sorted(fts.search("测试数据")) == [2]
    assert sorted(fts.search("world")) == [7]  # 只有文档 7 仍然包含 "world"
    
    # 测试批量删除的性能优势 - 同时删除多个文档
    large_batch = []
    large_ids = list(range(100, 150))
    
    for i in range(50):
        large_batch.append({
            "title": f"Batch Doc {i}",
            "content": f"批量文档内容 {i}",
            "tags": f"tag{i}"
        })
    
    # 批量添加
    fts.add_document(large_ids, large_batch)
    fts.flush()
    
    # 验证添加成功
    assert len(fts.search("批量文档内容")) == 50
    
    # 批量删除一半文档
    fts.remove_document(large_ids[:25])
    fts.flush()

    # 验证删除成功
    assert len(fts.search("批量文档内容")) == 25  # 只剩下一半文档
    
    # 验证特定文档已删除 - 使用tag搜索更精确
    for i in range(25):
        assert len(fts.search(f"tag{i}")) == 0  # 前25个文档的tag应该已被删除
    
    # 验证剩余文档仍可搜索 - 使用tag搜索更精确
    for i in range(25, 50):
        assert len(fts.search(f"tag{i}")) == 1  # 后25个文档的tag应该仍然存在



import re
import msgpack
from pathlib import Path
from collections import defaultdict
from typing import Dict, Union, Optional, List, Tuple

from pyroaring import BitMap
import xxhash

from .base import BaseIndex
from .file_manager import get_file_manager, safe_read_shard, safe_write_shard, safe_delete_shard


class InvertedIndex(BaseIndex):
    """Inverted index implementation"""
    
    def __init__(self, 
                 index_dir: Optional[Path] = None,
                 max_chinese_length: int = 4,
                 min_term_length: int = 2,
                 buffer_size: int = 100000,
                 shard_bits: int = 8,
                 cache_size: int = 1000,
                 fuzzy_threshold: float = 0.7,
                 fuzzy_max_distance: int = 2):
        """
        Initialize the inverted index
        
        Args:
            index_dir: The directory to store the index files
            max_chinese_length: The maximum length of Chinese substrings
            min_term_length: The minimum length of terms
            buffer_size: The size of the memory buffer
            shard_bits: The number of bits for the shard
            cache_size: The size of the cache
            fuzzy_threshold: The similarity threshold for fuzzy search (0.0-1.0)
            fuzzy_max_distance: The maximum edit distance for fuzzy search
        """
        self.index_dir = index_dir
        self.max_chinese_length = max_chinese_length
        self.min_term_length = min_term_length
        self.buffer_size = buffer_size
        
        self.chinese_pattern = re.compile(r'[\u4e00-\u9fff]+')
        self.modified_keys = set()
        
        # Main index and buffer
        self.word_index = defaultdict(BitMap)
        self.index_buffer = defaultdict(BitMap)
        
        # Chinese processing cache
        self._global_chinese_cache = {}
        
        self.shard_bits = shard_bits
        self.shard_count = 1 << shard_bits
        self.cache_size = cache_size
        
        # Fuzzy search parameters
        self.fuzzy_threshold = fuzzy_threshold
        self.fuzzy_max_distance = fuzzy_max_distance
        self._fuzzy_cache = {}  # Cache for fuzzy search results
        self._term_cache = None  # Cache for all terms, built on demand
        
        self._init_cache()

    def _init_cache(self) -> None:
        """Initialize the optimized cache"""
        # 使用普通字典代替LRU缓存，减少查找开销
        self._fast_cache = {}
        self._cache_hits = 0
        self._cache_max_size = self.cache_size
        
        def get_bitmap_fast(term: str) -> Optional[BitMap]:
            # 快速缓存查找
            if term in self._fast_cache:
                self._cache_hits += 1
                return self._fast_cache[term]
            
            # 首先检查缓冲区 - 避免不必要的拷贝
            if term in self.index_buffer:
                result = self.index_buffer[term]
                # 缓存结果（不拷贝，直接引用）
                self._cache_bitmap(term, result)
                return result
            
            # 检查内存索引（内存模式）
            if term in self.word_index:
                result = self.word_index[term] 
                self._cache_bitmap(term, result)
                return result
            
            # 如果是磁盘模式，从磁盘加载
            if self.index_dir:
                shard_id = self._get_shard_id(term)
                result = self._load_term_bitmap(term, shard_id)
                if result is not None:
                    self._cache_bitmap(term, result)
                return result
            
            return None
        
        self._bitmap_cache = get_bitmap_fast
    
    def _cache_bitmap(self, term: str, bitmap: BitMap) -> None:
        """缓存bitmap，管理缓存大小"""
        if len(self._fast_cache) >= self._cache_max_size:
            # 清理一半缓存（简单策略）
            keys_to_remove = list(self._fast_cache.keys())[:self._cache_max_size // 2]
            for key in keys_to_remove:
                del self._fast_cache[key]
        
        self._fast_cache[term] = bitmap

    def _get_shard_id(self, term: str) -> int:
        """Calculate the shard ID of the term"""
        return xxhash.xxh32(term.encode()).intdigest() & (self.shard_count - 1)

    def _load_term_bitmap(self, term: str, shard_id: int) -> Optional[BitMap]:
        """Load the bitmap of the term from the disk"""
        if not self.index_dir:
            return None
            
        shard_path = self.index_dir / 'shards' / f'shard_{shard_id}.nfts'
        if not shard_path.exists():
            return None
            
        try:
            fm = get_file_manager()
            with fm.safe_file_operation(shard_path, 'rb') as f:
                meta_size = int.from_bytes(f.read(4), 'big')
                meta_data = f.read(meta_size)
                meta = msgpack.unpackb(meta_data, raw=False)
                
                if term not in meta:
                    return None
                    
                offset, size = meta[term]
                f.seek(4 + meta_size + offset)
                bitmap_data = f.read(size)
                return BitMap.deserialize(bitmap_data)
        except:
            return None

    def add_terms(self, doc_id: int, terms: Dict[str, Union[str, int, float]]) -> None:
        """Add the document terms to the index"""
        for field_value in terms.values():
            field_str = str(field_value).lower()
            
            # Process the full field
            if len(field_str) >= self.min_term_length:
                self.index_buffer[field_str].add(doc_id)
            
            # Process Chinese
            for match in self.chinese_pattern.finditer(field_str):
                seg = match.group()
                if seg not in self._global_chinese_cache:
                    n = len(seg)
                    substrings = {seg[j:j + length] 
                                for length in range(self.min_term_length, 
                                                  min(n + 1, self.max_chinese_length + 1))
                                for j in range(n - length + 1)}
                    self._global_chinese_cache[seg] = substrings
                
                for substr in self._global_chinese_cache[seg]:
                    self.index_buffer[substr].add(doc_id)
            
            # Process English parts in mixed Chinese-English text
            english_parts = re.findall(r'[a-zA-Z]+', field_str)
            for eng_part in english_parts:
                if len(eng_part) >= self.min_term_length:
                    self.index_buffer[eng_part.lower()].add(doc_id)
            
            # Process the phrase
            if ' ' in field_str:
                self.index_buffer[field_str].add(doc_id)
                words = field_str.split()
                for word in words:
                    # 处理纯英文单词
                    if len(word) >= self.min_term_length and not self.chinese_pattern.search(word):
                        self.index_buffer[word].add(doc_id)
                    # 处理中英文混合单词，提取英文部分
                    elif self.chinese_pattern.search(word):
                        # 提取英文字母序列
                        english_parts = re.findall(r'[a-zA-Z]+', word)
                        for eng_part in english_parts:
                            if len(eng_part) >= self.min_term_length:
                                self.index_buffer[eng_part.lower()].add(doc_id)

        # Merge the buffer when it reaches the threshold
        if len(self.index_buffer) >= self.buffer_size:
            self.merge_buffer()
        
        # 无效化词条缓存，因为添加了新词条
        self._invalidate_term_cache()

    @staticmethod
    def _levenshtein_distance(s1: str, s2: str) -> int:
        """计算两个字符串的编辑距离（Levenshtein距离）"""
        if len(s1) < len(s2):
            s1, s2 = s2, s1
        
        if len(s2) == 0:
            return len(s1)
        
        previous_row = list(range(len(s2) + 1))
        for i, c1 in enumerate(s1):
            current_row = [i + 1]
            for j, c2 in enumerate(s2):
                insertions = previous_row[j + 1] + 1
                deletions = current_row[j] + 1
                substitutions = previous_row[j] + (c1 != c2)
                current_row.append(min(insertions, deletions, substitutions))
            previous_row = current_row
        
        return previous_row[-1]

    @staticmethod
    def _similarity_score(s1: str, s2: str) -> float:
        """计算两个字符串的相似度分数 (0.0-1.0)，针对中文优化"""
        if not s1 and not s2:
            return 1.0
        if not s1 or not s2:
            return 0.0
        
        # 基于编辑距离的相似度
        distance = InvertedIndex._levenshtein_distance(s1, s2)
        max_len = max(len(s1), len(s2))
        edit_similarity = 1.0 - (distance / max_len)
        
        # 基于字符重叠的相似度 (特别适合中文)
        set1, set2 = set(s1), set(s2)
        intersection = len(set1 & set2)
        union = len(set1 | set2)
        char_overlap = intersection / union if union > 0 else 0.0
        
        # 基于最长公共子序列的相似度
        def lcs_length(s1, s2):
            m, n = len(s1), len(s2)
            dp = [[0] * (n + 1) for _ in range(m + 1)]
            
            for i in range(1, m + 1):
                for j in range(1, n + 1):
                    if s1[i-1] == s2[j-1]:
                        dp[i][j] = dp[i-1][j-1] + 1
                    else:
                        dp[i][j] = max(dp[i-1][j], dp[i][j-1])
            return dp[m][n]
        
        lcs_len = lcs_length(s1, s2)
        lcs_similarity = lcs_len / max_len if max_len > 0 else 0.0
        
        # 混合相似度：取三种算法的最大值，更宽松地匹配中文
        # 这样"支持"和"一支"就能有合理的相似度了
        final_similarity = max(edit_similarity, char_overlap, lcs_similarity)
        
        return final_similarity

    def _get_all_terms(self) -> set:
        """获取所有索引中的词条，用于模糊搜索"""
        if self._term_cache is not None:
            return self._term_cache
        
        terms = set()
        
        # 从缓冲区获取词条
        terms.update(self.index_buffer.keys())
        
        # 从内存索引获取词条（内存模式）
        terms.update(self.word_index.keys())
        
        # 从磁盘分片获取词条（如果存在）
        if self.index_dir and (self.index_dir / 'shards').exists():
            shards_dir = self.index_dir / 'shards'
            for shard_path in shards_dir.glob('*.nfts'):
                try:
                    with open(shard_path, 'rb') as f:
                        meta_size = int.from_bytes(f.read(4), 'big')
                        meta_data = f.read(meta_size)
                        meta = msgpack.unpackb(meta_data, raw=False)
                        terms.update(meta.keys())
                except:
                    continue
        
        # 缓存结果
        self._term_cache = terms
        return terms

    def _invalidate_term_cache(self):
        """无效化词条缓存，在索引更新时调用"""
        self._term_cache = None

    def _fuzzy_search_terms(self, query: str, max_results: int = 20) -> List[Tuple[str, float]]:
        """在所有词条中进行模糊搜索，返回相似词条和相似度分数"""
        cache_key = f"{query}:{max_results}"
        if cache_key in self._fuzzy_cache:
            return self._fuzzy_cache[cache_key]
        
        all_terms = self._get_all_terms()
        candidates = []
        
        # 过滤长度相近的词条以提高效率
        query_len = len(query)
        max_len_diff = self.fuzzy_max_distance
        
        for term in all_terms:
            # 跳过长度差异过大的词条
            if abs(len(term) - query_len) > max_len_diff:
                continue
            
            # 快速预过滤：如果完全没有共同字符，跳过
            if not set(query) & set(term):
                continue
            
            similarity = self._similarity_score(query, term)
            if similarity >= self.fuzzy_threshold:
                candidates.append((term, similarity))
        
        # 按相似度降序排序，取前max_results个
        candidates.sort(key=lambda x: x[1], reverse=True)
        result = candidates[:max_results]
        
        # 缓存结果（限制缓存大小）
        if len(self._fuzzy_cache) > 1000:
            # 清理最旧的一半缓存
            oldest_keys = list(self._fuzzy_cache.keys())[:500]
            for key in oldest_keys:
                del self._fuzzy_cache[key]
        
        self._fuzzy_cache[cache_key] = result
        return result

    def search(self, query: str, enable_fuzzy: bool = False, min_results: int = 5) -> BitMap:
        """Search for a query - optimized version"""
        # 预处理优化：减少字符串操作
        if not query:
            return BitMap()
        
        # 只在必要时进行字符串转换
        if query != query.strip().lower():
            query = query.strip().lower()
            if not query:
                return BitMap()
        
        # 快速路径：单个英文单词（最常见的查询）
        if ' ' not in query and not self.chinese_pattern.search(query):
            result = self._bitmap_cache(query)
            if result is not None:
                # 对于读取操作，直接返回引用而不是拷贝
                return result
            else:
                return BitMap()  # 快速返回空结果
        
        # 直接匹配优化：避免不必要的拷贝
        result = self._bitmap_cache(query)
        if result is not None:
            # 对于读取操作，直接返回引用而不是拷贝
            return result
        
        # 2. Phrase query optimization - 支持中英文混合搜索
        if ' ' in query:
            words = query.split()
            if not words:
                return BitMap()
            
            # Pre-allocate the list to avoid dynamic growth
            results = []
            min_size = float('inf')
            min_idx = 0
            
            # 优化：预分配结果列表，提前检查词长度
            valid_words = [word for word in words if len(word) >= self.min_term_length]
            if not valid_words:
                return BitMap()
            
            # Get all the document sets of all the words at once
            for i, word in enumerate(valid_words):
                # 首先尝试直接匹配完整单词
                docs = self._bitmap_cache(word)
                
                # 如果直接匹配失败且单词包含中文，则尝试中文子串匹配
                if docs is None and self.chinese_pattern.search(word):
                    docs = self._search_chinese_word(word)
                
                if docs is None or len(docs) == 0:
                    return BitMap()  # Quick failure
                
                size = len(docs)
                if size < min_size:
                    min_size = size
                    min_idx = len(results)
                
                results.append(docs)
            
            if not results:
                return BitMap()
            
            # Optimization: directly use the smallest result set
            if min_idx > 0:
                results[0], results[min_idx] = results[min_idx], results[0]
            
            # Quick path: only one word
            if len(results) == 1:
                return results[0]
            
            # Efficient intersection
            result = results[0]
            for other in results[1:]:
                result &= other
                if not result:  # Return empty result early
                    return BitMap()
            
            return result
        
        # 3. Chinese query optimization
        exact_result = None
        if self.chinese_pattern.search(query):
            n = len(query)
            if n < self.min_term_length:
                return BitMap()
            
            # Optimization: directly use the longest possible substring
            max_len = min(n, self.max_chinese_length)
            
            # 3.1 Try the longest match
            for i in range(n - max_len + 1):
                substr = query[i:i + max_len]
                result = self._bitmap_cache(substr)
                if result is not None:
                    if len(result) < 1000:  # Return the result when it is small
                        exact_result = result
                        break
                    # Save the first match result
                    first_match = result
                    
                    # 3.2 Try to intersect with the adjacent substring
                    if i > 0:
                        prev_substr = query[i-1:i-1+max_len]
                        prev_docs = self._bitmap_cache(prev_substr)
                        if prev_docs is not None:
                            temp = result & prev_docs
                            if temp:
                                exact_result = temp
                                break
                    
                    if i < n - max_len:
                        next_substr = query[i+1:i+1+max_len]
                        next_docs = self._bitmap_cache(next_substr)
                        if next_docs is not None:
                            temp = result & next_docs
                            if temp:
                                exact_result = temp
                                break
                    
                    exact_result = first_match
                    break
            
            # 3.3 Fall back to the minimum length match
            if exact_result is None:
                for i in range(n - self.min_term_length + 1):
                    substr = query[i:i + self.min_term_length]
                    result = self._bitmap_cache(substr)
                    if result is not None:
                        exact_result = result
                        break
        
        # 4. 模糊搜索（仅在启用且精确搜索结果不足时）
        if enable_fuzzy and (exact_result is None or len(exact_result) < min_results):
            fuzzy_result = self._perform_fuzzy_search(query)
            
            if exact_result is None:
                return fuzzy_result
            elif len(exact_result) < min_results and len(fuzzy_result) > 0:
                # 合并精确搜索和模糊搜索的结果
                return exact_result | fuzzy_result
        
        return exact_result if exact_result is not None else BitMap()

    def _search_chinese_word(self, word: str) -> Optional[BitMap]:
        """对包含中文的单词进行子串搜索"""
        n = len(word)
        if n < self.min_term_length:
            return None
        
        # 尝试不同长度的子串，从最长开始
        max_len = min(n, self.max_chinese_length)
        
        # 首先尝试最长匹配
        for i in range(n - max_len + 1):
            substr = word[i:i + max_len]
            result = self._bitmap_cache(substr)
            if result is not None:
                if len(result) < 1000:  # 如果结果较少，直接返回
                    return result
                # 保存第一个匹配结果
                first_match = result
                
                # 尝试与相邻子串交集以获得更精确的结果
                if i > 0:
                    prev_substr = word[i-1:i-1+max_len]
                    prev_docs = self._bitmap_cache(prev_substr)
                    if prev_docs is not None:
                        temp = result & prev_docs
                        if temp:
                            return temp
                
                if i < n - max_len:
                    next_substr = word[i+1:i+1+max_len]
                    next_docs = self._bitmap_cache(next_substr)
                    if next_docs is not None:
                        temp = result & next_docs
                        if temp:
                            return temp
                
                return first_match
        
        # 如果最长匹配失败，尝试最小长度匹配
        for i in range(n - self.min_term_length + 1):
            substr = word[i:i + self.min_term_length]
            result = self._bitmap_cache(substr)
            if result is not None:
                return result
        
        return None

    def _perform_fuzzy_search(self, query: str) -> BitMap:
        """执行模糊搜索并返回合并的结果"""
        similar_terms = self._fuzzy_search_terms(query)
        if not similar_terms:
            return BitMap()
        
        # 合并所有相似词条的结果
        result = BitMap()
        for term, similarity in similar_terms:
            term_result = self._bitmap_cache(term)
            if term_result is not None:
                # 可以根据相似度加权，但为了性能简单起见，直接合并
                result |= term_result
        
        return result

    def remove_document(self, doc_id: int) -> None:
        """Remove the document from the index"""
        if not self.index_dir:
            return
        
        # Remove from the buffer
        keys_to_remove = []
        for key, doc_ids in self.index_buffer.items():
            if doc_id in doc_ids:
                doc_ids.discard(doc_id)
                self.modified_keys.add(key)
                if not doc_ids:
                    keys_to_remove.append(key)
        
        for key in keys_to_remove:
            del self.index_buffer[key]
        
        # Remove from the shards
        shards_dir = self.index_dir / 'shards'
        if shards_dir.exists():
            fm = get_file_manager()
            for shard_path in shards_dir.glob('*.nfts'):
                try:
                    # Read the shard data using safe file operations
                    meta, shard_data = safe_read_shard(shard_path)
                    
                    modified = False
                    new_data = {}
                    new_meta = {}
                    offset = 0
                    
                    # Process each term
                    for term, bitmap_data in shard_data.items():
                        bitmap = BitMap.deserialize(bitmap_data)
                        
                        if doc_id in bitmap:
                            bitmap.discard(doc_id)
                            modified = True
                            self.modified_keys.add(term)
                            
                        if bitmap:  # Only save non-empty bitmap
                            bitmap_data = bitmap.serialize()
                            new_data[term] = bitmap_data
                            new_meta[term] = (offset, len(bitmap_data))
                            offset += len(bitmap_data)
                    
                    # If there are modifications, rewrite the shard file
                    if modified:
                        if new_data:
                            safe_write_shard(shard_path, new_meta, new_data)
                        else:
                            # If the shard is empty, delete the file
                            safe_delete_shard(shard_path)
                            
                except Exception as e:
                    print(f"Error processing shard {shard_path}: {e}")
                    continue
        
        # Remove from the word index
        keys_to_remove = []
        for key, doc_ids in self.word_index.items():
            if doc_id in doc_ids:
                doc_ids.discard(doc_id)
                self.modified_keys.add(key)
                if not doc_ids:
                    keys_to_remove.append(key)
        
        for key in keys_to_remove:
            del self.word_index[key]
        
        # Clear the cache
        self._fast_cache.clear()
        
        # Ensure buffer changes are immediately available for search
        # This is especially important on Windows platforms
        self.merge_buffer()

    def update_terms(self, doc_id: int, terms: Dict[str, Union[str, int, float]]) -> None:
        """Update the terms of the document"""
        # 获取文档当前的所有词条
        current_terms = set()
        
        # 从缓冲区中获取
        for term, bitmap in self.index_buffer.items():
            if doc_id in bitmap:
                current_terms.add(term)
                
        # 从分片文件中获取
        if self.index_dir:
            shards_dir = self.index_dir / 'shards'
            if shards_dir.exists():
                for shard_path in shards_dir.glob('*.nfts'):
                    try:
                        with open(shard_path, 'rb') as f:
                            meta_size = int.from_bytes(f.read(4), 'big')
                            meta_data = f.read(meta_size)
                            meta = msgpack.unpackb(meta_data, raw=False)
                            
                            for term, (offset, size) in meta.items():
                                f.seek(4 + meta_size + offset)
                                bitmap_data = f.read(size)
                                bitmap = BitMap.deserialize(bitmap_data)
                                if doc_id in bitmap:
                                    current_terms.add(term)
                    except Exception as e:
                        print(f"Error reading shard {shard_path}: {e}")
                        continue
        
        # 从词索引中获取
        for term, bitmap in self.word_index.items():
            if doc_id in bitmap:
                current_terms.add(term)
        
        # 生成新的词条集合
        new_terms = set()
        for field_value in terms.values():
            field_str = str(field_value).lower()
            
            # 处理完整字段
            if len(field_str) >= self.min_term_length:
                new_terms.add(field_str)
            
            # 处理中文
            for match in self.chinese_pattern.finditer(field_str):
                seg = match.group()
                if seg not in self._global_chinese_cache:
                    n = len(seg)
                    substrings = {seg[j:j + length] 
                                for length in range(self.min_term_length, 
                                                  min(n + 1, self.max_chinese_length + 1))
                                for j in range(n - length + 1)}
                    self._global_chinese_cache[seg] = substrings
                
                new_terms.update(self._global_chinese_cache[seg])
            
            # 处理短语
            if ' ' in field_str:
                new_terms.add(field_str)
                words = field_str.split()
                for word in words:
                    if len(word) >= self.min_term_length and not self.chinese_pattern.search(word):
                        new_terms.add(word)
        
        # 找出需要删除和添加的词条
        terms_to_remove = current_terms - new_terms
        terms_to_add = new_terms - current_terms
        
        # 删除旧词条
        for term in terms_to_remove:
            # 从缓冲区删除
            if term in self.index_buffer:
                self.index_buffer[term].discard(doc_id)
                self.modified_keys.add(term)
                if not self.index_buffer[term]:
                    del self.index_buffer[term]
            
            # 从词索引中删除
            if term in self.word_index:
                self.word_index[term].discard(doc_id)
                self.modified_keys.add(term)
                if not self.word_index[term]:
                    del self.word_index[term]
            
            # 从分片文件中删除
            if self.index_dir:
                shard_id = self._get_shard_id(term)
                shard_path = self.index_dir / 'shards' / f'shard_{shard_id}.nfts'
                if shard_path.exists():
                    try:
                        meta, shard_data = safe_read_shard(shard_path)
                        
                        if term in shard_data:
                            bitmap = BitMap.deserialize(shard_data[term])
                            bitmap.discard(doc_id)
                            self.modified_keys.add(term)
                            
                            # 更新分片文件
                            if bitmap:
                                new_data = {term: bitmap.serialize()}
                                new_meta = {term: (0, len(new_data[term]))}
                                safe_write_shard(shard_path, new_meta, new_data)
                            else:
                                # 如果位图为空，删除文件
                                safe_delete_shard(shard_path)
                    except Exception as e:
                        print(f"Error updating shard {shard_path}: {e}")
                        continue
        
        # 添加新词条
        for term in terms_to_add:
            self.index_buffer[term].add(doc_id)
            self.modified_keys.add(term)
            
            # 如果是词条，也添加到词索引中
            if ' ' in term:
                words = term.split()
                for word in words:
                    if len(word) >= self.min_term_length and not self.chinese_pattern.search(word):
                        self.word_index[word].add(doc_id)
                        self.modified_keys.add(word)
        
        # 强制合并缓冲区以确保更新立即生效
        # 这对于Windows平台特别重要
        self.merge_buffer()
        
        # 清理缓存
        self._fast_cache.clear()

    def batch_update_terms(self, doc_ids: List[int], docs_terms: List[Dict[str, Union[str, int, float]]]) -> None:
        """批量更新多个文档的词条
        
        Args:
            doc_ids: 文档ID列表
            docs_terms: 文档词条列表，与doc_ids一一对应
        """
        if len(doc_ids) != len(docs_terms):
            raise ValueError("文档ID列表和文档词条列表长度不匹配")
            
        if not doc_ids:
            return
            
        # 收集所有文档的当前词条
        all_current_terms = defaultdict(set)  # term -> set(doc_ids)
        all_new_terms = defaultdict(set)      # term -> set(doc_ids)
        
        # 1. 收集所有文档的当前词条
        # 从缓冲区收集
        for term, bitmap in self.index_buffer.items():
            for doc_id in doc_ids:
                if doc_id in bitmap:
                    all_current_terms[term].add(doc_id)
        
        # 从分片文件收集
        if self.index_dir:
            shards_dir = self.index_dir / 'shards'
            if shards_dir.exists():
                for shard_path in shards_dir.glob('*.nfts'):
                    try:
                        with open(shard_path, 'rb') as f:
                            meta_size = int.from_bytes(f.read(4), 'big')
                            meta_data = f.read(meta_size)
                            meta = msgpack.unpackb(meta_data, raw=False)
                            
                            for term, (offset, size) in meta.items():
                                f.seek(4 + meta_size + offset)
                                bitmap_data = f.read(size)
                                bitmap = BitMap.deserialize(bitmap_data)
                                
                                for doc_id in doc_ids:
                                    if doc_id in bitmap:
                                        all_current_terms[term].add(doc_id)
                    except Exception as e:
                        print(f"Error reading shard {shard_path}: {e}")
                        continue
        
        # 从词索引收集
        for term, bitmap in self.word_index.items():
            for doc_id in doc_ids:
                if doc_id in bitmap:
                    all_current_terms[term].add(doc_id)
        
        # 2. 生成所有文档的新词条
        for i, (doc_id, terms) in enumerate(zip(doc_ids, docs_terms)):
            # 处理每个文档的词条
            for field_value in terms.values():
                field_str = str(field_value).lower()
                
                # 处理完整字段
                if len(field_str) >= self.min_term_length:
                    all_new_terms[field_str].add(doc_id)
                
                # 处理中文
                for match in self.chinese_pattern.finditer(field_str):
                    seg = match.group()
                    if seg not in self._global_chinese_cache:
                        n = len(seg)
                        substrings = {seg[j:j + length] 
                                    for length in range(self.min_term_length, 
                                                      min(n + 1, self.max_chinese_length + 1))
                                    for j in range(n - length + 1)}
                        self._global_chinese_cache[seg] = substrings
                    
                    for substr in self._global_chinese_cache[seg]:
                        all_new_terms[substr].add(doc_id)
                
                # 处理短语
                if ' ' in field_str:
                    all_new_terms[field_str].add(doc_id)
                    words = field_str.split()
                    for word in words:
                        if len(word) >= self.min_term_length and not self.chinese_pattern.search(word):
                            all_new_terms[word].add(doc_id)
        
        # 3. 计算需要更新的词条
        # 对于每个词条，找出需要删除和添加的文档ID
        terms_to_update = set(all_current_terms.keys()) | set(all_new_terms.keys())
        
        # 按分片分组，减少文件I/O
        sharded_updates = defaultdict(lambda: defaultdict(lambda: {'add': set(), 'remove': set()}))
        
        for term in terms_to_update:
            current_docs = all_current_terms.get(term, set())
            new_docs = all_new_terms.get(term, set())
            
            # 需要删除的文档ID
            docs_to_remove = current_docs - new_docs
            # 需要添加的文档ID
            docs_to_add = new_docs - current_docs
            
            if not docs_to_remove and not docs_to_add:
                continue  # 没有变化，跳过
                
            # 更新缓冲区
            if term in self.index_buffer:
                for doc_id in docs_to_remove:
                    self.index_buffer[term].discard(doc_id)
                for doc_id in docs_to_add:
                    self.index_buffer[term].add(doc_id)
                if not self.index_buffer[term]:
                    del self.index_buffer[term]
            else:
                for doc_id in docs_to_add:
                    self.index_buffer[term].add(doc_id)
            
            # 更新词索引
            if ' ' in term:
                words = term.split()
                for word in words:
                    if len(word) >= self.min_term_length and not self.chinese_pattern.search(word):
                        if word in self.word_index:
                            for doc_id in docs_to_remove:
                                self.word_index[word].discard(doc_id)
                            for doc_id in docs_to_add:
                                self.word_index[word].add(doc_id)
                            if not self.word_index[word]:
                                del self.word_index[word]
                        else:
                            for doc_id in docs_to_add:
                                self.word_index[word].add(doc_id)
            
            # 按分片分组
            if self.index_dir:
                shard_id = self._get_shard_id(term)
                sharded_updates[shard_id][term]['remove'].update(docs_to_remove)
                sharded_updates[shard_id][term]['add'].update(docs_to_add)
            
            # 标记为已修改
            self.modified_keys.add(term)
        
        # 4. 更新分片文件
        if self.index_dir:
            for shard_id, terms in sharded_updates.items():
                shard_path = self.index_dir / 'shards' / f'shard_{shard_id}.nfts'
                if not shard_path.exists() and not any(terms[term]['add'] for term in terms):
                    continue  # 没有需要添加的文档，且分片不存在，跳过
                
                # 读取现有分片数据
                existing_meta = {}
                existing_data = {}
                
                if shard_path.exists():
                    try:
                        existing_meta, shard_data = safe_read_shard(shard_path)
                        
                        for term, bitmap_data in shard_data.items():
                            if term in terms:  # 只更新需要更新的词条
                                bitmap = BitMap.deserialize(bitmap_data)
                                
                                # 更新位图
                                for doc_id in terms[term]['remove']:
                                    bitmap.discard(doc_id)
                                for doc_id in terms[term]['add']:
                                    bitmap.add(doc_id)
                                
                                if bitmap:
                                    existing_data[term] = bitmap.serialize()
                            else:
                                existing_data[term] = bitmap_data
                    except Exception as e:
                        print(f"Error reading shard {shard_path}: {e}")
                        continue
                
                # 添加新词条
                for term in terms:
                    if term not in existing_data and terms[term]['add']:
                        bitmap = BitMap(terms[term]['add'])
                        existing_data[term] = bitmap.serialize()
                
                # 删除空位图
                for term in list(existing_data.keys()):
                    if term in terms and not terms[term]['add'] and term in existing_meta:
                        bitmap = BitMap.deserialize(existing_data[term])
                        for doc_id in terms[term]['remove']:
                            bitmap.discard(doc_id)
                        if not bitmap:
                            del existing_data[term]
                
                # 保存更新后的分片
                if existing_data:
                    # 重建元数据
                    new_meta = {}
                    offset = 0
                    for term, data in existing_data.items():
                        new_meta[term] = (offset, len(data))
                        offset += len(data)
                    
                    # 使用安全写入
                    safe_write_shard(shard_path, new_meta, existing_data)
                elif shard_path.exists():
                    # 如果没有数据，安全删除文件
                    safe_delete_shard(shard_path)
        
        # 5. 强制合并缓冲区以确保批量更新立即生效
        # 这对于Windows平台特别重要
        self.merge_buffer()
        
        # 6. 清理缓存
        self._fast_cache.clear()

    def merge_buffer(self) -> None:
        """Merge the buffer"""
        if not self.index_buffer:
            return
        
        # 如果是内存模式（index_dir为None），直接合并到内存索引
        if not self.index_dir:
            for term, bitmap in self.index_buffer.items():
                if term in self.word_index:
                    self.word_index[term] |= bitmap
                else:
                    self.word_index[term] = bitmap.copy()
            
            # Clear the buffer
            self.index_buffer.clear()
            # Clear the cache
            self._fast_cache.clear()
            # 无效化词条缓存
            self._invalidate_term_cache()
            return
            
        # 磁盘模式：分片处理
        # Group by shard
        sharded_buffer = defaultdict(lambda: defaultdict(BitMap))
        for term, bitmap in self.index_buffer.items():
            shard_id = self._get_shard_id(term)
            sharded_buffer[shard_id][term] = bitmap
        
        # Write to shards
        for shard_id, terms in sharded_buffer.items():
            self._merge_shard(shard_id, terms)
        
        # Clear the buffer
        self.index_buffer.clear()
        # Clear the cache
        self._fast_cache.clear()
        # 无效化词条缓存
        self._invalidate_term_cache()

    def _merge_shard(self, shard_id: int, terms: Dict[str, BitMap]) -> None:
        """Merge the data of a single shard"""
        shard_path = self.index_dir / 'shards' / f'shard_{shard_id}.nfts'
        
        # Read the existing data
        existing_meta = {}
        existing_data = {}
        if shard_path.exists():
            try:
                existing_meta, shard_data = safe_read_shard(shard_path)
                existing_data = shard_data
            except Exception as e:
                print(f"Error reading shard {shard_path} for merge: {e}")
                existing_data = {}
        
        # Merge the data
        new_data = {}
        new_meta = {}
        offset = 0
        
        for term, bitmap_data in existing_data.items():
            if term in terms:  # Need to update
                bitmap = BitMap.deserialize(bitmap_data)
                bitmap |= terms[term]
                bitmap_data = bitmap.serialize()
                del terms[term]
            
            new_data[term] = bitmap_data
            new_meta[term] = (offset, len(bitmap_data))
            offset += len(bitmap_data)
        
        # Add new terms
        for term, bitmap in terms.items():
            bitmap_data = bitmap.serialize()
            new_data[term] = bitmap_data
            new_meta[term] = (offset, len(bitmap_data))
            offset += len(bitmap_data)
        
        # Save using safe operations
        if new_data:
            safe_write_shard(shard_path, new_meta, new_data)

    def save(self, incremental: bool = True) -> None:
        """Save the index"""
        if not self.index_dir:
            return
            
        # Save the buffer data by shard
        if self.index_buffer:
            self.merge_buffer()
        
        # Save the word index
        if self.word_index:
            word_dir = self.index_dir / 'word'
            word_dir.mkdir(exist_ok=True)
            self._save_shard(self.word_index, word_dir / "index.nfts", incremental)
        
        if incremental:
            self.modified_keys.clear()

    def _save_shard(self, shard_data: Dict[str, BitMap], shard_path: Path, incremental: bool) -> None:
        """Save a single shard
        
        Args:
            shard_data: The shard data
            shard_path: The shard path
            incremental: Whether to save incrementally
        """
        if not shard_data:
            if shard_path.exists():
                safe_delete_shard(shard_path)
            return
            
        # Process the incremental update
        existing_meta = {}
        existing_data = {}
        if incremental and shard_path.exists():
            try:
                existing_meta, existing_shard_data = safe_read_shard(shard_path)
                
                for key, bitmap_data in existing_shard_data.items():
                    if key not in self.modified_keys:
                        existing_data[key] = bitmap_data
            except Exception as e:
                print(f"Error reading existing shard {shard_path}: {e}")
                # Continue with empty existing data
        
        # Prepare new data
        data = {}
        meta = {}
        offset = 0
        
        # Process the existing data
        for key, bitmap_data in existing_data.items():
            meta[key] = (offset, len(bitmap_data))
            data[key] = bitmap_data
            offset += len(bitmap_data)
        
        # Process the new data
        for key, bitmap in shard_data.items():
            if not bitmap:
                continue
            if not incremental or key in self.modified_keys:
                bitmap_data = bitmap.serialize()
                data[key] = bitmap_data
                meta[key] = (offset, len(bitmap_data))
                offset += len(bitmap_data)
        
        # Save using safe operations
        if data:
            safe_write_shard(shard_path, meta, data)
        elif shard_path.exists():
            safe_delete_shard(shard_path)

    def load(self) -> bool:
        """Load the index"""
        if not self.index_dir:
            return False
            
        try:
            # Load shards
            shards_dir = self.index_dir / 'shards'
            if not shards_dir.exists():
                return False
                
            for shard_path in shards_dir.glob('*.nfts'):
                self._load_shard(shard_path)
            
            # Load the word index
            word_dir = self.index_dir / 'word'
            if word_dir.exists():
                word_index_path = word_dir / "index.nfts"
                if word_index_path.exists():
                    self._load_shard(word_index_path, is_word_index=True)
            
            return True
        except Exception as e:
            print(f"Failed to load the index: {e}")
            return False

    def _load_shard(self, shard_path: Path, is_word_index: bool = False) -> None:
        """Load a single shard
        
        Args:
            shard_path: The shard path
            is_word_index: Whether it is the word index
        """
        if not shard_path.exists():
            return
            
        try:
            meta, shard_data = safe_read_shard(shard_path)
            
            for key, bitmap_data in shard_data.items():
                if len(key) >= self.min_term_length:
                    bitmap = BitMap.deserialize(bitmap_data)
                    if is_word_index:
                        self.word_index[key] |= bitmap
                    else:
                        # Add the data to the buffer
                        self.index_buffer[key] |= bitmap
        
            # If it is not the word index, merge to the disk
            if not is_word_index and self.index_buffer:
                self.merge_buffer()
                
        except Exception as e:
            print(f"Failed to load the shard {shard_path}: {e}")

    def build_word_index(self) -> None:
        """Build the word index"""
        if not self.index_dir:
            return
        
        temp_word_index = defaultdict(set)
        shards_dir = self.index_dir / 'shards'
        
        # Read the data from all shards
        if shards_dir.exists():
            for shard_path in shards_dir.glob('*.nfts'):
                try:
                    meta, shard_data = safe_read_shard(shard_path)
                    
                    for term, bitmap_data in shard_data.items():
                        if ' ' in term:  # Only process the word with space
                            bitmap = BitMap.deserialize(bitmap_data)
                            
                            # Process the words in the phrase
                            words = term.split()
                            doc_ids = list(bitmap)
                            for word in words:
                                if not self.chinese_pattern.search(word) and len(word) >= self.min_term_length:
                                    temp_word_index[word].update(doc_ids)
                except Exception as e:
                    print(f"Failed to process the shard {shard_path}: {e}")
                    continue
        
        # Process the data in the buffer
        for term, bitmap in self.index_buffer.items():
            if ' ' in term:
                words = term.split()
                doc_ids = list(bitmap)
                for word in words:
                    if not self.chinese_pattern.search(word) and len(word) >= self.min_term_length:
                        temp_word_index[word].update(doc_ids)
        
        # Convert to BitMap and save
        self.word_index.clear()
        for word, doc_ids in temp_word_index.items():
            if doc_ids:
                self.word_index[word] = BitMap(doc_ids)
        
        # Save the word index
        if self.index_dir:
            word_dir = self.index_dir / 'word'
            word_dir.mkdir(exist_ok=True)
            self._save_shard(self.word_index, word_dir / "index.nfts", False) 

    def batch_remove_document(self, doc_ids: List[int]) -> None:
        """批量删除多个文档
        
        Args:
            doc_ids: 要删除的文档ID列表
        """
        if not doc_ids:
            return
            
        if not self.index_dir:
            return
            
        # 从缓冲区中删除
        keys_to_remove = []
        for key, bitmap in self.index_buffer.items():
            modified = False
            for doc_id in doc_ids:
                if doc_id in bitmap:
                    bitmap.discard(doc_id)
                    modified = True
            
            if modified:
                self.modified_keys.add(key)
                if not bitmap:
                    keys_to_remove.append(key)
        
        for key in keys_to_remove:
            del self.index_buffer[key]
        
        # 从分片文件中删除
        shards_dir = self.index_dir / 'shards'
        if shards_dir.exists():
            # 按分片分组，减少文件I/O
            sharded_updates = defaultdict(set)
            
            # 收集所有需要更新的分片
            for shard_path in shards_dir.glob('*.nfts'):
                try:
                    meta, shard_data = safe_read_shard(shard_path)
                    
                    modified = False
                    new_data = {}
                    new_meta = {}
                    offset = 0
                    
                    # 处理每个词条
                    for term, bitmap_data in shard_data.items():
                        bitmap = BitMap.deserialize(bitmap_data)
                        
                        term_modified = False
                        for doc_id in doc_ids:
                            if doc_id in bitmap:
                                bitmap.discard(doc_id)
                                term_modified = True
                        
                        if term_modified:
                            modified = True
                            self.modified_keys.add(term)
                            
                        if bitmap:  # 只保存非空位图
                            bitmap_data = bitmap.serialize()
                            new_data[term] = bitmap_data
                            new_meta[term] = (offset, len(bitmap_data))
                            offset += len(bitmap_data)
                    
                    # 如果有修改，重写分片文件
                    if modified:
                        if new_data:
                            safe_write_shard(shard_path, new_meta, new_data)
                        else:
                            # 如果分片为空，安全删除文件
                            safe_delete_shard(shard_path)
                            
                except Exception as e:
                    print(f"Error processing shard {shard_path}: {e}")
                    continue
        
        # 从词索引中删除
        keys_to_remove = []
        for key, bitmap in self.word_index.items():
            modified = False
            for doc_id in doc_ids:
                if doc_id in bitmap:
                    bitmap.discard(doc_id)
                    modified = True
            
            if modified:
                self.modified_keys.add(key)
                if not bitmap:
                    keys_to_remove.append(key)
        
        for key in keys_to_remove:
            del self.word_index[key]
        
        # 清理缓存
        self._fast_cache.clear()
        
        # 强制合并缓冲区以确保批量删除立即生效
        # 这对于Windows平台特别重要
        self.merge_buffer() 


"""
平台特定的文件句柄管理器
专门处理Windows上的文件句柄未及时释放问题
"""

import platform
import time
import gc
from typing import Optional, Callable, Any
from pathlib import Path
import contextlib


class FileHandleManager:
    """平台特定的文件句柄管理器"""
    
    def __init__(self):
        self.is_windows = platform.system() == 'Windows'
        self.retry_count = 3 if self.is_windows else 1
        self.retry_delay = 0.01  # 10ms delay between retries
    
    @contextlib.contextmanager
    def safe_file_operation(self, file_path: Path, mode: str = 'rb'):
        """安全的文件操作上下文管理器
        
        Args:
            file_path: 文件路径
            mode: 文件打开模式
            
        Yields:
            文件对象
        """
        file_obj = None
        for attempt in range(self.retry_count):
            try:
                file_obj = open(file_path, mode)
                yield file_obj
                break
            except (OSError, IOError, PermissionError) as e:
                if file_obj:
                    try:
                        file_obj.close()
                    except:
                        pass
                    file_obj = None
                
                if attempt == self.retry_count - 1:
                    raise e
                
                # Windows特定的句柄释放策略
                if self.is_windows:
                    self._force_handle_release()
                    time.sleep(self.retry_delay * (attempt + 1))
            finally:
                if file_obj:
                    try:
                        file_obj.close()
                    except:
                        pass
                    
                # Windows特定的后处理
                if self.is_windows:
                    self._post_operation_cleanup()
    
    def safe_file_delete(self, file_path: Path) -> bool:
        """安全删除文件
        
        Args:
            file_path: 要删除的文件路径
            
        Returns:
            是否删除成功
        """
        if not file_path.exists():
            return True
            
        for attempt in range(self.retry_count):
            try:
                file_path.unlink()
                return True
            except (OSError, IOError, PermissionError) as e:
                if attempt == self.retry_count - 1:
                    # 在Windows上，如果删除失败，至少记录错误但不抛出异常
                    if self.is_windows:
                        print(f"Warning: Could not delete file {file_path}: {e}")
                        return False
                    else:
                        raise e
                
                # Windows特定的句柄释放策略
                if self.is_windows:
                    self._force_handle_release()
                    time.sleep(self.retry_delay * (attempt + 1))
        
        return False
    
    def _force_handle_release(self):
        """强制释放文件句柄（Windows特定）"""
        if self.is_windows:
            # 强制垃圾回收，释放未关闭的文件句柄
            gc.collect()
            # 额外的短暂等待，让操作系统释放句柄
            time.sleep(0.001)  # 1ms
    
    def _post_operation_cleanup(self):
        """操作后清理（Windows特定）"""
        if self.is_windows:
            # 强制垃圾回收
            gc.collect()


# 全局文件管理器实例
_file_manager = None

def get_file_manager() -> FileHandleManager:
    """获取全局文件管理器实例"""
    global _file_manager
    if _file_manager is None:
        _file_manager = FileHandleManager()
    return _file_manager


def safe_read_shard(file_path: Path) -> tuple[dict, dict]:
    """安全读取分片文件
    
    Args:
        file_path: 分片文件路径
        
    Returns:
        (meta_data, shard_data) 元组
    """
    import msgpack
    from pyroaring import BitMap
    
    fm = get_file_manager()
    
    with fm.safe_file_operation(file_path, 'rb') as f:
        meta_size = int.from_bytes(f.read(4), 'big')
        meta_data = f.read(meta_size)
        meta = msgpack.unpackb(meta_data, raw=False)
        
        shard_data = {}
        for term, (offset, size) in meta.items():
            f.seek(4 + meta_size + offset)
            bitmap_data = f.read(size)
            shard_data[term] = bitmap_data
    
    return meta, shard_data


def safe_write_shard(file_path: Path, meta: dict, data: dict) -> bool:
    """安全写入分片文件
    
    Args:
        file_path: 分片文件路径
        meta: 元数据
        data: 分片数据
        
    Returns:
        是否写入成功
    """
    import msgpack
    
    fm = get_file_manager()
    
    try:
        file_path.parent.mkdir(exist_ok=True, parents=True)
        
        with fm.safe_file_operation(file_path, 'wb') as f:
            meta_data = msgpack.packb(meta, use_bin_type=True)
            f.write(len(meta_data).to_bytes(4, 'big'))
            f.write(meta_data)
            for bitmap_data in data.values():
                f.write(bitmap_data)
        
        return True
    except Exception as e:
        print(f"Error writing shard {file_path}: {e}")
        return False


def safe_delete_shard(file_path: Path) -> bool:
    """安全删除分片文件
    
    Args:
        file_path: 分片文件路径
        
    Returns:
        是否删除成功
    """
    fm = get_file_manager()
    return fm.safe_file_delete(file_path)


import re
import shutil
from typing import List, Union, Dict
from pathlib import Path

from pyroaring import BitMap

from .factory import IndexFactory, IndexType
from .inserter import DocumentInserter


class FullTextSearch:
    def __init__(self, 
                 index_dir: str = None, 
                 max_chinese_length: int = 4, 
                 num_workers: int = 8,
                 shard_size: int = 500_000,
                 min_term_length: int = 2,
                 auto_save: bool = True,
                 batch_size: int = 10000,
                 drop_if_exists: bool = False,
                 buffer_size: int = 100000,
                 fuzzy_threshold: float = 0.4,
                 fuzzy_max_distance: int = 2):
        """
        Initialize the full-text search engine

        Args:
            index_dir (str): The directory to store the index files, if None, use in-memory index
            max_chinese_length (int): The maximum length of Chinese substrings, default is 4
            num_workers (int): The number of worker threads for parallel indexing, default is 8
            shard_size (int): The number of documents per shard, default is 500,000
            min_term_length (int): The minimum length of a term, default is 2
            auto_save (bool): Whether to save to disk automatically, default is True
            batch_size (int): The number of documents to process in each batch, default is 10000
            drop_if_exists (bool): Whether to delete the index files if they exist, default is False
            buffer_size (int): The size of the memory buffer, default is 100000
            fuzzy_threshold (float): The similarity threshold for fuzzy search (0.0-1.0), default is 0.7
            fuzzy_max_distance (int): The maximum edit distance for fuzzy search, default is 2
        """
        self.chinese_pattern = re.compile(r'[\u4e00-\u9fff]+')
        self.index_dir = Path(index_dir) if index_dir else None
        
        if drop_if_exists and self.index_dir and self.index_dir.exists():
            shutil.rmtree(self.index_dir)
                
        self.max_chinese_length = max_chinese_length
        self.num_workers = num_workers
        self.shard_size = shard_size
        self.min_term_length = min_term_length
        self.auto_save = auto_save
        self.batch_size = batch_size
        self.buffer_size = buffer_size
        self.fuzzy_threshold = fuzzy_threshold
        self.fuzzy_max_distance = fuzzy_max_distance
        
        index_type_enum = IndexType("inverted")
            
        self.inverted_index = IndexFactory.create_index(
            index_type=index_type_enum,
            index_dir=index_dir,
            max_chinese_length=max_chinese_length,
            min_term_length=min_term_length,
            buffer_size=buffer_size,
            fuzzy_threshold=fuzzy_threshold,
            fuzzy_max_distance=fuzzy_max_distance
        )
        
        self.inserter = DocumentInserter(
            index=self.inverted_index,
            num_workers=num_workers,
            batch_size=batch_size,
            shard_size=shard_size
        )
        
        self._batch_count = 0
        
        if self.index_dir:
            self.index_dir.mkdir(parents=True, exist_ok=True)
            self.inverted_index.load()

    def add_document(self, doc_id: Union[int, List[int]], 
                    fields: Union[Dict[str, Union[str, int, float]], 
                                List[Dict[str, Union[str, int, float]]]]):
        """添加一个或多个文档到索引
        
        Args:
            doc_id: 单个文档ID或文档ID列表
            fields: 单个文档的字段或文档字段列表，与doc_id一一对应
            
        Examples:
            # 添加单个文档
            fts.add_document(1, {"title": "doc1", "content": "content1"})
            
            # 批量添加多个文档
            fts.add_document([1, 2], [
                {"title": "doc1", "content": "content1"},
                {"title": "doc2", "content": "content2"}
            ])
        """
        self.inserter.add_documents(doc_id, fields)

    def search(self, query: str, enable_fuzzy: bool = False, min_results: int = 5) -> Union[BitMap, List[tuple[int, float]]]:
        """Search for a query
        
        Args:
            query: The query to search for
            enable_fuzzy: Whether to enable fuzzy search, default is False
            min_results: Minimum number of results before enabling fuzzy search, default is 5
            
        Returns:
            Union[BitMap, List[tuple[int, float]]]: The document ID set or a list of (document ID, similarity)
            
        Examples:
            # 精确搜索
            results = fts.search("搜索查询")
            
            # 启用模糊搜索
            results = fts.search("搜索查询", enable_fuzzy=True)
            
            # 设置最小结果数阈值
            results = fts.search("搜索查询", enable_fuzzy=True, min_results=10)
        """
        return self.inverted_index.search(query, enable_fuzzy=enable_fuzzy, min_results=min_results)

    def fuzzy_search(self, query: str, min_results: int = 5) -> Union[BitMap, List[tuple[int, float]]]:
        """便捷的模糊搜索方法
        
        Args:
            query: The query to search for
            min_results: Minimum number of results before enabling fuzzy search, default is 5
            
        Returns:
            Union[BitMap, List[tuple[int, float]]]: The document ID set or a list of (document ID, similarity)
        """
        return self.search(query, enable_fuzzy=True, min_results=min_results)

    def flush(self):
        """Flush the buffer and save to disk"""
        self.inserter.flush()

    def remove_document(self, doc_id: Union[int, List[int]]):
        """从索引中删除一个或多个文档
        
        Args:
            doc_id: 单个文档ID或要删除的文档ID列表
            
        Examples:
            # 删除单个文档
            fts.remove_document(1)
            
            # 批量删除多个文档
            fts.remove_document([1, 2, 3])
        """
        if isinstance(doc_id, int):
            self.inverted_index.remove_document(doc_id)
        else:
            self.inverted_index.batch_remove_document(doc_id)
            
        if self.index_dir:
            self.inverted_index.save(incremental=True)

    def update_document(self, doc_id: Union[int, List[int]], 
                       fields: Union[Dict[str, Union[str, int, float]], 
                                  List[Dict[str, Union[str, int, float]]]]):
        """更新一个或多个文档
        
        Args:
            doc_id: 单个文档ID或文档ID列表
            fields: 单个文档的更新字段或文档字段列表，与doc_id一一对应
            
        Examples:
            # 更新单个文档
            fts.update_document(1, {"title": "new title", "content": "new content"})
            
            # 批量更新多个文档
            fts.update_document([1, 2], [
                {"title": "new title 1", "content": "new content 1"},
                {"title": "new title 2", "content": "new content 2"}
            ])
        """
        if isinstance(doc_id, int):
            self.inserter.update_document(doc_id, fields)
        else:
            self.inserter.batch_update_document(doc_id, fields)

    def from_pandas(self, df, id_column=None, text_columns=None):
        """Import data from a pandas DataFrame
        
        Args:
            df: pandas DataFrame object
            id_column: The name of the document ID column, if None, use the row index
            text_columns: The list of text columns to index, if None, use all string columns
        """
        self.inserter.from_pandas(df, id_column, text_columns)

    def from_polars(self, df, id_column=None, text_columns=None):
        """Import data from a polars DataFrame
        
        Args:
            df: polars DataFrame object
            id_column: The name of the document ID column, if None, use the row index
            text_columns: The list of text columns to index, if None, use all string columns
        """
        self.inserter.from_polars(df, id_column, text_columns)

    def from_arrow(self, table, id_column=None, text_columns=None):
        """Import data from a pyarrow Table
        
        Args:
            table: pyarrow Table object
            id_column: The name of the document ID column, if None, use the row index
            text_columns: The list of text columns to index, if None, use all string columns
        """
        self.inserter.from_arrow(table, id_column, text_columns)

    def from_parquet(self, path, id_column=None, text_columns=None):
        """Import data from a parquet file
        
        Args:
            path: The path to the parquet file
            id_column: The name of the document ID column, if None, use the row index
            text_columns: The list of text columns to index, if None, use all string columns
        """
        self.inserter.from_parquet(path, id_column, text_columns)

    def from_csv(self, path, id_column=None, text_columns=None):
        """Import data from a CSV file
        
        Args:
            path: The path to the CSV file
            id_column: The name of the document ID column, if None, use the row index
            text_columns: The list of text columns to index, if None, use all string columns
        """
        self.inserter.from_csv(path, id_column, text_columns)

    def get_fuzzy_config(self) -> Dict[str, Union[float, int]]:
        """获取模糊搜索配置
        
        Returns:
            Dict containing fuzzy search configuration
        """
        return {
            "fuzzy_threshold": self.fuzzy_threshold,
            "fuzzy_max_distance": self.fuzzy_max_distance
        }

    def set_fuzzy_config(self, fuzzy_threshold: float = None, fuzzy_max_distance: int = None):
        """设置模糊搜索配置
        
        Args:
            fuzzy_threshold: The similarity threshold for fuzzy search (0.0-1.0)
            fuzzy_max_distance: The maximum edit distance for fuzzy search
        """
        if fuzzy_threshold is not None:
            self.fuzzy_threshold = fuzzy_threshold
            self.inverted_index.fuzzy_threshold = fuzzy_threshold
            
        if fuzzy_max_distance is not None:
            self.fuzzy_max_distance = fuzzy_max_distance
            self.inverted_index.fuzzy_max_distance = fuzzy_max_distance

